DISPLAY_NAME = "Beckman Echo Plate Reformat"
DEVICE_TYPE = "liquid handler"
PRODUCT_MANUFACTURER = "Beckman Coulter"
SOFTWARE_NAME = "Labcyte Echo Plate Reformat"

TRANSFER_RESULTS_LABEL = "DETAILS"
TRANSFER_ERRORS_LABEL = "EXCEPTIONS"

PLATE_REFORMAT_REPORT_VOLUME_UNITS = "nL"
PLATE_REFORMAT_REPORT_VOLUME_CONVERSION_TO_UL = 0.001
