DISPLAY_NAME = "Cytiva Biacore Insight"
DEVICE_IDENTIFIER = "Biacore"
DEVICE_TYPE = "Binding Affinity Analyzer"
DETECTION_TYPE = "Surface Plasmon Resonance"
PRODUCT_MANUFACTURER = "Cytiva"

COMPARTMENT_TEMP_REGEX = r"Sample compartment temperature (\d+),.*"
